/*
 * Videology gspca Camera driver
 *
 * Supports:
 *		Videology 60V002 Interface Board	id:[0x0143]
 *
 * Copyright (C) 2009, 2010 Brian Maly <brian_maly@yahoo.com>
 * 
 * V4L2 by Jean-Francois Moine <http://moinejf.free.fr>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Based on gspca / gspca_videology_interlaced
 */

#include <linux/version.h>


/* more optimal to use the gspca.h from your dist kernel, if installed */
//#include "/usr/src/kernels/2.6.31.5-127.fc12.x86_64/include/config/usb/gspca.h"

/* otherwise use local gspca header until its shipped by default with the kernel src */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,39)
#include "gspca-2.6.39.h"
#elif LINUX_VERSION_CODE == KERNEL_VERSION(2,6,38)
#include "gspca-2.6.38.h"
#elif LINUX_VERSION_CODE == KERNEL_VERSION(2,6,37)
#include "gspca-2.6.37.h"
#elif LINUX_VERSION_CODE == KERNEL_VERSION(2,6,36)
#include "gspca-2.6.36.h"
#elif LINUX_VERSION_CODE == KERNEL_VERSION(2,6,35)
#include "gspca-2.6.35.h"
#elif LINUX_VERSION_CODE == KERNEL_VERSION(2,6,34)
#include "gspca-2.6.34.h"
#elif LINUX_VERSION_CODE == KERNEL_VERSION(2,6,33)
#include "gspca-2.6.33.h"
#elif LINUX_VERSION_CODE == KERNEL_VERSION(2,6,32)
#include "gspca-2.6.32.h"
#else
#include "gspca-2.6.31.h"
#endif


#define MODULE_NAME "gspca_videology_framegrabber"

MODULE_AUTHOR("Brian Maly <brian_maly@yahoo.com>");
MODULE_DESCRIPTION("Videology USB Interface Board v0.17");
MODULE_LICENSE("GPL");

#define USB_VIDEOLOGY_VENDOR_ID         0x1bbd
#define USB_VIDEOLOGY_PRODUCT_ID_FG	0x0143

#define SENSOR_NTSC	0x42
#define SENSOR_PAL	0x43

/* Firmware command registers (controller) */
#define VX_RESET_FIFO   0xBB
#define VX_I2C_CMD      0x00C3
#define VX_CAMERA_POWER 0xC4
#define VX_SET_PORT_D   0xC8
#define VX_GET_SNAPSHOT 0xC9

/* Camera registers (CCD) */
#define SENSOR		0x00
#define ANALOG_GAIN	0x05
#define COLOR_MODE	0x08
#define BRIGHTNESS	0x0a
#define CONTRAST	0x0b
#define SATURATION	0x0c
#define HUE		0x0d
#define CHROMA		0x0f
#define D_RESTORE	0x51
#define V_SIGNAL	0x70
#define V_STANDARD	0x71

/* Devices (cameras) */
#define CAM_ADDR 0x0068

/* Subdevices (camera modules) */
#define EEPROM_PAGE_1_ADDR 0x0020

/* Sensors type codes */
#define SENSOR_FG	0xff

#define VIDEOLOGY_CTRL_TIMEOUT		100  /* ms */

 /* Mod param 	*/
static int frame_offset;
static int default_mode;

/* specific webcam descriptor */
struct sd {
	struct gspca_dev gspca_dev;		/* !! must be the first item */
	
	/* standard V4L2 controls */	
	u8 mode;
	u8 videosignalsource;
	u8 gain;
 	u8 contrast;
	u8 brightness;
	u8 saturation;
	u8 chromagain;
	u8 hue;

	/* driver internal */
	int framesize;
	int sof;	
};

/* V4L2 controls supported */
static int sd_querymenu(struct gspca_dev *gspca_dev, struct v4l2_querymenu *menu);
static int sd_setgain(struct gspca_dev *gspca_dev, __s32 val);
static int sd_getgain(struct gspca_dev *gspca_dev, __s32 *val);
static int sd_setchromagain(struct gspca_dev *gspca_dev, s32 val);
static int sd_getchromagain(struct gspca_dev *gspca_dev, __s32 *val);
static int sd_setsaturation(struct gspca_dev *gspca_dev, s32 val);
static int sd_getsaturation(struct gspca_dev *gspca_dev, __s32 *val);
static int sd_setvideosignalsource(struct gspca_dev *gspca_dev, s32 val);
static int sd_getvideosignalsource(struct gspca_dev *gspca_dev, __s32 *val);
static int sd_setcontrast(struct gspca_dev *gspca_dev, __s32 val);
static int sd_getcontrast(struct gspca_dev *gspca_dev, __s32 *val);
static int sd_setbrightness(struct gspca_dev *gspca_dev, __s32 val);
static int sd_getbrightness(struct gspca_dev *gspca_dev, __s32 *val);
static int sd_sethue(struct gspca_dev *gspca_dev, __s32 val);
static int sd_gethue(struct gspca_dev *gspca_dev, __s32 *val);
static int videology_setup_camera_pal_mode(struct gspca_dev *gspca_dev);
static int videology_setup_camera_ntsc_mode(struct gspca_dev *gspca_dev);
static int videology_setup_camera_alt_ntsc_mode(struct gspca_dev *gspca_dev);

static struct ctrl sd_ctrls[] = {
    
#define HUE_IDX 0
        {
            {
                .id      = V4L2_CID_HUE,
                .type    = V4L2_CTRL_TYPE_INTEGER,
                .name    = "Hue",
                .minimum = 0,
                .maximum = 0x7f,
                .step    = 1,
#define HUE_DEFAULT 0x00
                .default_value = HUE_DEFAULT,
            }, 
            .set = sd_sethue,
            .get = sd_gethue,
        },

#define CHROMA_IDX 1
        {
            {
                .id      = V4L2_CID_CHROMA_AGC,  
                .type    = V4L2_CTRL_TYPE_INTEGER,
                .name    = "Chroma Gain",
                .minimum = 0,
                .maximum = 0xff,
                .step    = 1,
#define CHROMAGAIN_DEFAULT 0x2a
                .default_value = CHROMAGAIN_DEFAULT,
            },
            .set = sd_setchromagain,
            .get = sd_getchromagain,
        },

#define CONTRAST_IDX 2
        {
            {
                .id      = V4L2_CID_CONTRAST,
                .type    = V4L2_CTRL_TYPE_INTEGER,
                .name    = "Contrast",
                .minimum = 0,
                .maximum = 0x7f,
                .step    = 1,
#define CONTRAST_DEFAULT 0x47
                .default_value = CONTRAST_DEFAULT,
            },
            .set = sd_setcontrast,
            .get = sd_getcontrast,
        },

#define BRIGHTNESS_IDX 3
        {
            {
                .id      = V4L2_CID_BRIGHTNESS,
                .type    = V4L2_CTRL_TYPE_INTEGER,
                .name    = "Brightness",
                .minimum = 0,
                .maximum = 0xff,
                .step    = 1,
#define BRIGHTNESS_DEFAULT 0x80
                .default_value = BRIGHTNESS_DEFAULT,
            },
            .set = sd_setbrightness,
            .get = sd_getbrightness,
        },

#define SATURATION_IDX 4
        {
            {
                .id      = V4L2_CID_SATURATION,
                .type    = V4L2_CTRL_TYPE_INTEGER,
                .name    = "Saturation",
                .minimum = 0,
               	.maximum = 0x7f,
                .step    = 1,
#define SATURATION_DEFAULT 0x40
                .default_value = SATURATION_DEFAULT,
            },
            .set = sd_setsaturation,
            .get = sd_getsaturation,
        },

#define GAIN_IDX 5
	{
            {
                .id      = V4L2_CID_GAIN,
                .type    = V4L2_CTRL_TYPE_INTEGER,
                .name    = "Analog Input Gain",
                .minimum = 0x00,
                .maximum = 0xff, 
                .step    = 1,
#define GAIN_DEFAULT 0x81
                .default_value = GAIN_DEFAULT,
            },
            .set = sd_setgain,
            .get = sd_getgain,
        },

#define VIDEO_SIGNAL_SOURCE_IDX 6
        {
            {
                .id      = V4L2_CID_USER_CLASS,
                .type    = V4L2_CTRL_TYPE_MENU,
                .name    = "Video Signal Source",
                .minimum = 0,
                .maximum = 1,
                .step    = 1,
#define VIDEO_SIGNAL_SOURCE_DEFAULT 0
                .default_value = VIDEO_SIGNAL_SOURCE_DEFAULT,
            }, 
            .set = sd_setvideosignalsource,
            .get = sd_getvideosignalsource,
        },

};

/* Format sizes */
/* PAL */
#define PAL_NUM_ACTIVE_PIXELS 724 /* 720+SAV+EAV=724 */
#define PAL_NUM_ACTIVE_LINES 243//244 /* 486/2 = 243 for interlaced */
/* NTSC (standard) */
#define NTSC_NUM_ACTIVE_PIXELS 724 /* 720+SAV+EAV=724 */
#define NTSC_NUM_ACTIVE_LINES 241 /* 480/2 + 1/2 = 241 for interlaced */
/* NTSC (alternate size) */
#define NTSC_ALT_NUM_ACTIVE_PIXELS 724 /* 720+SAV+EAV=724 */
#define NTSC_ALT_NUM_ACTIVE_LINES 288 /* 576/2 = 288 for interlaced */

static const struct v4l2_pix_format cam_mode[] = {
        {PAL_NUM_ACTIVE_PIXELS, PAL_NUM_ACTIVE_LINES, V4L2_PIX_FMT_UYVY, V4L2_FIELD_SEQ_TB,
		.bytesperline = PAL_NUM_ACTIVE_PIXELS * 2,
                .sizeimage = (PAL_NUM_ACTIVE_PIXELS * 2) * PAL_NUM_ACTIVE_LINES, 
		.colorspace = V4L2_COLORSPACE_SMPTE170M,
                .priv = 0,                     
	},
        {NTSC_NUM_ACTIVE_PIXELS, NTSC_NUM_ACTIVE_LINES, V4L2_PIX_FMT_UYVY, V4L2_FIELD_SEQ_TB,
		.bytesperline = NTSC_NUM_ACTIVE_PIXELS * 2,
                .sizeimage = (NTSC_NUM_ACTIVE_PIXELS * 2) * NTSC_NUM_ACTIVE_LINES, 
		.colorspace = V4L2_COLORSPACE_SMPTE170M,
                .priv = 1,                     
	},
	{NTSC_ALT_NUM_ACTIVE_PIXELS, NTSC_ALT_NUM_ACTIVE_LINES, V4L2_PIX_FMT_UYVY, V4L2_FIELD_SEQ_TB,
		.bytesperline = NTSC_ALT_NUM_ACTIVE_PIXELS * 2,
                .sizeimage = (NTSC_ALT_NUM_ACTIVE_PIXELS * 2) * NTSC_ALT_NUM_ACTIVE_LINES, 
		.colorspace = V4L2_COLORSPACE_SMPTE170M,
                .priv = 2,                     
	},

};

static int cam_register_write(struct gspca_dev *gspca_dev, u8 request, u16 value, u16 index)
{
	struct usb_device *udev = gspca_dev->dev;
	int err;

	//printk("cam_register_write(0x%x, 0x%x)\n", request, value);
	err = usb_control_msg(udev, usb_sndctrlpipe(udev, 0), request,
			USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_INTERFACE,
			value, index, NULL, 0, VIDEOLOGY_CTRL_TIMEOUT);
	
	if (err)
		printk("Failed to write a control register (index 0x%04X,"
			" value 0x%02X, error %d)\n", index, value, err);

	return err;
}

static int cam_register_read(struct gspca_dev *gspca_dev, u16 req, u16 val)
{
	struct usb_device *udev = gspca_dev->dev;
	int res;
	u8 j[4];

	//printk("cam_register_read(0x%x,0x%x)\n", req, val);			
	res = usb_control_msg(udev, usb_rcvctrlpipe(udev, 0), 
			req /* req */,
			USB_DIR_IN | USB_TYPE_VENDOR | USB_RECIP_INTERFACE,
                        val /* value */, 
			0 /* index */, 
			j, 
			3, VIDEOLOGY_CTRL_TIMEOUT);


	if (res < 0) {
		printk("Failed to read a register (reg 0x%x, error %d)\n",
			val, res);
		return res;
	}
	
//	printk("read [0x%x 0x%x 0x%x]\n", j[0], j[1], j[2]);
	return j[1];
}

static int i2c_register_write(struct gspca_dev *gspca_dev, int device, int subdevice, int val)
{
//	printk("i2c_register_write(device 0x%x, subdevice 0x%x, val 0x%x)\n", device, subdevice, val);
	cam_register_write(gspca_dev, VX_I2C_CMD, 0x0300 | device, 0);		/* device address */
	cam_register_write(gspca_dev, VX_I2C_CMD, 0x0200 | subdevice, 0);	/* subdevice address */
	cam_register_write(gspca_dev, VX_I2C_CMD, 0x0600 | val, 0); 		/* value to write */
	return 0;
}

static int i2c_register_read(struct gspca_dev *gspca_dev, int device, int subdevice)
{
	u8 r;

	//printk("i2c_register_read(device 0x%x, subvdevice 0x%x)\n", device, subdevice);
	cam_register_write(gspca_dev, VX_I2C_CMD, 0x0300 | device, 0);		/* device address */
	cam_register_write(gspca_dev, VX_I2C_CMD, 0x0200 | subdevice, 0);	/* subdevice address */
	cam_register_write(gspca_dev, VX_I2C_CMD, 0x0300 | device | 0x01, 0); 	/* device addr | access mode */
	
	r = cam_register_read(gspca_dev, VX_I2C_CMD, 0x6ff);
	//printk("i2c_register_read(0x%x) register = 0x%x\n", subdevice, r);

	return r;
}

static int set_altsetting(struct usb_interface *iface, int alternate)
{
        struct usb_device               *dev;
	//printk("set_altsetting(%d)\n", alternate);
        if (alternate < 0 || alternate >= 256)
                return -EINVAL;

        dev = interface_to_usbdev (iface);
        return usb_set_interface (dev,
                        iface->altsetting [0].desc.bInterfaceNumber,
                        alternate);
}

static int videology_dump_camera_registers(struct gspca_dev *gspca_dev)
{
	u8 r;

	/* dump the value of every reg we have */
	printk("----------------- regs --------------------------\n");

	r = i2c_register_read(gspca_dev, CAM_ADDR, SENSOR);
	printk("reg SENSOR = 0x%x\n", r); 

	r = i2c_register_read(gspca_dev, CAM_ADDR, 0x01);
	printk("reg 0x01 = 0x%x\n", r); 

	r = i2c_register_read(gspca_dev, CAM_ADDR, 0x02);
	printk("reg 0x02 = 0x%x\n", r); 
	
	r = i2c_register_read(gspca_dev, CAM_ADDR, 0x03);
	printk("reg 0x03 = 0x%x\n", r);

	r = i2c_register_read(gspca_dev, CAM_ADDR, 0x04); 
	printk("reg 0x04 = 0x%x\n", r);	

	r = i2c_register_read(gspca_dev, CAM_ADDR, ANALOG_GAIN);
	printk("reg ANALOG_GAIN = 0x%x\n", r);  

	r = i2c_register_read(gspca_dev, CAM_ADDR, 0x06);  
	printk("reg 0x06 = 0x%x\n", r);

	r = i2c_register_read(gspca_dev, CAM_ADDR, 0x07);  
	printk("reg 0x07 = 0x%x\n", r);

	r = i2c_register_read(gspca_dev, CAM_ADDR, COLOR_MODE);  
	printk("reg COLOR_MODE = 0x%x\n", r);

	r = i2c_register_read(gspca_dev, CAM_ADDR, 0x09);  
	printk("reg 0x09 = 0x%x\n", r);

	r = i2c_register_read(gspca_dev, CAM_ADDR, BRIGHTNESS);  
	printk("reg BRIGHTNESS = 0x%x\n", r);

	r = i2c_register_read(gspca_dev, CAM_ADDR, CONTRAST);  
	printk("reg CONTRAST = 0x%x\n", r);

	r = i2c_register_read(gspca_dev, CAM_ADDR, SATURATION);  
	printk("reg SATURATION = 0x%x\n", r);

	r = i2c_register_read(gspca_dev, CAM_ADDR, HUE);  
	printk("reg HUE = 0x%x\n", r);

	r = i2c_register_read(gspca_dev, CAM_ADDR, 0x0e);  
	printk("reg 0x0e = 0x%x\n", r);

	r = i2c_register_read(gspca_dev, CAM_ADDR, CHROMA);  
	printk("reg CHROMA = 0x%x\n", r);

	r = i2c_register_read(gspca_dev, CAM_ADDR, 0x10);  
	printk("reg 0x10 = 0x%x\n", r);

	r = i2c_register_read(gspca_dev, CAM_ADDR, 0x11);  
	printk("reg 0x11 = 0x%x\n", r);

	r = i2c_register_read(gspca_dev, CAM_ADDR, 0x12);  
	printk("reg 0x12 = 0x%x\n", r);

	r = i2c_register_read(gspca_dev, CAM_ADDR, 0x13);  
	printk("reg 0x13 = 0x%x\n", r);

	r = i2c_register_read(gspca_dev, CAM_ADDR, V_SIGNAL);  
	printk("reg V_SIGNAL = 0x%x\n", r);

	r = i2c_register_read(gspca_dev, CAM_ADDR, V_STANDARD);  
	printk("reg V_STANDARD = 0x%x\n", r);

	printk("----------------- end regs ----------------------\n");

	return 0;
}
/* setup camera */
static int videology_setup_camera_defaults(struct gspca_dev *gspca_dev)
{
	struct sd *sd = (struct sd *) gspca_dev;
	u8 r; 
//	printk("videology_setup_camera()\n");

	printk("Videology interface board detected (60V002)\n");

	/* select our default video standard */
	
	videology_setup_camera_pal_mode(gspca_dev);
	msleep(250);

	/* read current camera settings and set v4l2 controls accordingly */

	r = i2c_register_read(gspca_dev, CAM_ADDR, V_SIGNAL);
	/* for some reason, if we replug the camera some stored values get altered */
	/* when we write 0x0, store, and read back the eeprom val we get 0x1 */
	/* and when we write 0x1, store, and read back from eeprom we get 0x3 */
	/* so here we superimpose the correct values to fix this */
	if (r == 0x1)
		sd->videosignalsource = 0;
	if (r == 0x3)
		sd->videosignalsource = 1;

	r = i2c_register_read(gspca_dev, CAM_ADDR, ANALOG_GAIN);
	if (r >= 0)
		sd->gain = r;

	r = i2c_register_read(gspca_dev, CAM_ADDR, BRIGHTNESS);
	if (r >= 0)
		sd->brightness = r;

	r = i2c_register_read(gspca_dev, CAM_ADDR, CONTRAST);
	if (r >= 0)
		sd->contrast = r;

	r = i2c_register_read(gspca_dev, CAM_ADDR, SATURATION);
	if (r >= 0)
		sd->saturation = r;

	r = i2c_register_read(gspca_dev, CAM_ADDR, HUE);
	if (r >= 0)
		sd->hue = r;

	r = i2c_register_read(gspca_dev, CAM_ADDR, CHROMA);
	if (r >= 0)
		sd->chromagain = r;
 
	return 0;
}

/* setup camera pal mode */
static int videology_setup_camera_pal_mode(struct gspca_dev *gspca_dev)
{
	struct cam *cam = &gspca_dev->cam;
//        printk("videology_setup_camera_pal_mode()\n");

	/* set PAL (default) mode for all cams */
  	cam->cam_mode = cam_mode;
	cam->nmodes = ARRAY_SIZE(cam_mode);
	gspca_dev->width = PAL_NUM_ACTIVE_PIXELS;
	gspca_dev->height = PAL_NUM_ACTIVE_LINES;
	//i2c_register_write(gspca_dev, CAM_ADDR, COLOR_MODE, 0x81);
	i2c_register_write(gspca_dev, CAM_ADDR, V_STANDARD, 0);
	msleep(250);

        return 0;
}

/* setup camera ntsc mode */
static int videology_setup_camera_ntsc_mode(struct gspca_dev *gspca_dev)
{
	struct cam *cam = &gspca_dev->cam;
//        printk("videology_setup_camera_ntsc_mode()\n");
	/* set NTSC (normal) mode for most cams */

	cam->cam_mode = cam_mode;
	cam->nmodes = ARRAY_SIZE(cam_mode);
	gspca_dev->width = NTSC_NUM_ACTIVE_PIXELS;
	gspca_dev->height = NTSC_NUM_ACTIVE_LINES;
	//i2c_register_write(gspca_dev, CAM_ADDR, COLOR_MODE, 0x40);
	i2c_register_write(gspca_dev, CAM_ADDR, V_STANDARD, 1);
	msleep(250);

        return 0;
}

/* setup camera alt ntsc mode */
static int videology_setup_camera_alt_ntsc_mode(struct gspca_dev *gspca_dev)
{
	struct cam *cam = &gspca_dev->cam;
//        printk("videology_setup_camera_alt_ntsc_mode()\n");

 	/* set NTSC 576 mode for blue cam */
	cam->cam_mode = cam_mode;
	cam->nmodes = ARRAY_SIZE(cam_mode);
	gspca_dev->width = NTSC_ALT_NUM_ACTIVE_PIXELS;
	gspca_dev->height = NTSC_ALT_NUM_ACTIVE_LINES;
	i2c_register_write(gspca_dev, CAM_ADDR, V_STANDARD, 1);

        return 0;
}
/* this function is called at probe time */
static int sd_config(struct gspca_dev *gspca_dev,
			const struct usb_device_id *id)
{
	struct sd *sd = (struct sd *) gspca_dev;
	struct cam *cam;
	struct usb_interface *intf;

//	printk("sd_config()\n");
	printk("Videology camera detected (vid/pid 0x%04X:0x%04X)\n", 
		id->idVendor, id->idProduct);

	intf = usb_ifnum_to_if(gspca_dev->dev, gspca_dev->iface);
	set_altsetting(intf, 1); 

	/* device specific config */
	cam = &gspca_dev->cam;
  	cam->cam_mode = cam_mode;
        cam->nmodes = ARRAY_SIZE(cam_mode);

	cam->bulk = 0; /* we use isoc */
	/* gspca_main tries to use the highest valid altconfig, however  */
	/* some videology cameras advertize altconfigs used only for debugging */
	/* as valid altconfigs. */
	/* unless we explicitly limit the altconfigs available, gspca_main will */
	/* pick a debugging altconfig assuming its perfectly valid */ 
	gspca_dev->nbalt = 2;

	/* gspca specific config */
	/* set all defaults */
	sd->gain = GAIN_DEFAULT;
	sd->videosignalsource = VIDEO_SIGNAL_SOURCE_DEFAULT;
        sd->saturation = SATURATION_DEFAULT;
        sd->chromagain = CHROMAGAIN_DEFAULT;
	sd->contrast = CONTRAST_DEFAULT;
	sd->brightness = BRIGHTNESS_DEFAULT;
	sd->hue = HUE_DEFAULT;

	return 0;
}

/* this function is called at probe and resume time */
static int sd_init(struct gspca_dev *gspca_dev)
{
//	printk("sd_init()\n");

	/* stop CLPD */
	//printk("stopping CLPD\n");
	cam_register_write(gspca_dev, VX_SET_PORT_D, 0x80, 0);
	cam_register_write(gspca_dev, VX_SET_PORT_D, 0x80, 0);

	/* camera power on */
	//printk("turning camera power on...\n");
	cam_register_write(gspca_dev, VX_CAMERA_POWER, 1, 0); 

	/* allow camera settle */
	mdelay(250);

	//printk("doing camera init...\n");
	videology_setup_camera_defaults(gspca_dev);

	return 0;
}

/* -- start the camera -- */
static int sd_start(struct gspca_dev *gspca_dev)
{
	struct sd *sd = (struct sd *) gspca_dev;
//	printk("sd_start()\n");
	struct usb_interface *intf;
	intf = usb_ifnum_to_if(sd->gspca_dev.dev, sd->gspca_dev.iface);

	/* double clear FIFO */
        cam_register_write(gspca_dev, VX_RESET_FIFO, 0, 0);
        cam_register_write(gspca_dev, VX_RESET_FIFO, 0, 0);
	
	/* camera dumps 0xff for all regs if read too soon after powerup */
	/* dump reg debug info in start() rather than probe() */
	//videology_dump_camera_registers(gspca_dev);

 	/* set our video mode to whatever the viewer is requesting */
        sd->mode = gspca_dev->cam.cam_mode[gspca_dev->curr_mode].priv;

	/* check to see if the user wants to manually override the */
	/* viewers current mode */
	if (default_mode > 0) 
		sd->mode = default_mode - 1;

	printk("sd->mode=%d\n", sd->mode);
	switch (sd->mode) {
                case 2:
                        videology_setup_camera_alt_ntsc_mode(gspca_dev);
                        break;
                case 1:
                        videology_setup_camera_ntsc_mode(gspca_dev);
                        break;
                case 0:
                default:
                        videology_setup_camera_pal_mode(gspca_dev);
                        break;
        }

	/* trigger */
	//printk("triggering...\n");	
        cam_register_write(gspca_dev, VX_SET_PORT_D, 0, 0);
	sd->sof = 0;

	return 0;
}


static void sd_stopN(struct gspca_dev *gspca_dev)
{
	//printk("sd_stopN()\n");	
}

static void sd_stop0(struct gspca_dev *gspca_dev)
{
	//printk("sd_stop0()\n");
	cam_register_write(gspca_dev, VX_SET_PORT_D, 0x80, 0);
	cam_register_write(gspca_dev, VX_SET_PORT_D, 0x80, 0);
}


static void  dq_callback(struct gspca_dev *gspca_dev)
{
	//printk("dq_callback()\n");
}

/* locate the end of header */
static int find_bt656_eoh(unsigned char *data, int mark, int len)
{
	int p; 

	if (mark == -1) 
		mark = 0;

	for (p = mark; p < (len - 4); p++) { 
                if ((data[p+0] == 0xff) && (data[p+1] == 0x00) && (data[p+2] == 0x00) && ((data[p+3] == 0x80) || (data[p+3] == 0xc6))) 
			return p;
	}
	return -1;
}

static int find_bt656_sof(unsigned char *data, int mark, int len)
{
	int p;

	/* check fov vertical blanking end bit set */
	/* return position of SOF */
	for (p = mark; p < (len - 4); p++) { 
                if ((data[p+0] == 0xff) && (data[p+1] == 0x00) && (data[p+2] == 0x00) && ((data[p+3] == 0xf0) /*|| (data[p+3] == 0xb6)*/ ))
			return p;
	}
	return -1;
}


/* frame depricated for kernels >= 2.6.33 */
/* use u8 instead of __u8 for >= 2.6.33 */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,33)
static void sd_pkt_scan(struct gspca_dev *gspca_dev,
			u8 *data,
			int len)
#else
static void sd_pkt_scan(struct gspca_dev *gspca_dev,
			struct gspca_frame *frame,
			__u8 *data,
			int len)
#endif
{
		struct sd *sd = (struct sd *) gspca_dev;
		int sof = -1;
		int eof = -1;
		int eoh = -1;
//printk("pkt_scan(len=%d)\n", len);

		/* find the start of frame marker */
		sof = find_bt656_sof(data, 0, len);

		eoh = find_bt656_eoh(data, 0, len);

		if (sof >= 0 || eoh >= 0) {
				
			if (sd->sof && sof >= 0 && eoh == -1) {	
				/* discard oversized frames */
				eof = sof;
				if (sd->sof && (sd->framesize + eof + frame_offset > (gspca_dev->height * (gspca_dev->width * 2))) ) {
					eof -= ((sd->framesize + eof + frame_offset) - (gspca_dev->height * (gspca_dev->width * 2)));
				}
				if (sd->sof && (sd->framesize + eof + frame_offset < (gspca_dev->height * (gspca_dev->width * 2))) ) {
					eof += ((gspca_dev->height * (gspca_dev->width * 2)) - (sd->framesize + eof + frame_offset)) /*- 1*/;
				}
				if (sd->sof && (sd->framesize + eof + frame_offset != (gspca_dev->height * (gspca_dev->width * 2))) ) {
					/* drop frame... incorrect size */
//					printk("DISCARD size %d should be %d\n", sd->framesize, (sd->framesize + eof + frame_offset));
					gspca_dev->last_packet_type = DISCARD_PACKET;
					sd->framesize = 0;
					sd->sof = 0;
					return;
				}

				/* add the end of frame up to the end of the EOF */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,33)

				gspca_frame_add(gspca_dev, LAST_PACKET, data, eof + frame_offset); 
#else
				frame = gspca_frame_add(gspca_dev, LAST_PACKET, frame, data, eof + frame_offset);
#endif
				sd->sof = 0;
				sd->framesize += (eof + frame_offset);
//				printk("LAST size = %d bytes 	lines = %d\n", sd->framesize, sd->framesize / (gspca_dev->width * 2));
				return;
			}
			if (!sd->sof && eoh >= 0 /*&& sof == -1*/) {

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,33)
				gspca_frame_add(gspca_dev, FIRST_PACKET, data + eoh + frame_offset, len - (eoh + frame_offset)); 
#else
				gspca_frame_add(gspca_dev, FIRST_PACKET, frame, data + eoh + frame_offset, len - (eoh + frame_offset));
#endif
//				printk("FIRST len %d     sof %d     eoh %d     code 0x%x\n", len, sof, eoh,  data[sof+4]);
				sd->sof = 1;
				sd->framesize = len - (eoh + frame_offset);
				return;
			}
		}
		if (sd->sof == 1 && len /*&& sof == -1 && eoh == -1*/) {

			if (sd->framesize + len > (gspca_dev->height * (gspca_dev->width * 2)) ) {
				/* handle gspca overflow */
//				printk("DISCARD size %d should be %d\n", sd->framesize, (sd->framesize + len));
				gspca_dev->last_packet_type = DISCARD_PACKET;	
				sd->sof = 0;
				sd->framesize = 0;
				return;
			}
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,33)
			gspca_frame_add(gspca_dev, INTER_PACKET, data, len);
#else
			/* use frame context for older kernels */
			gspca_frame_add(gspca_dev, INTER_PACKET, frame, data, len);
#endif
			sd->framesize += len;
//			printk("INTER sof %d eoh %d		len %d		framesize now %d\n", sof, eoh, len, sd->framesize);
			return;
		}
		else {
			/* we must have encountered an ISOC error and missed the header */
//			printk("DISCARD unknown [len %d] sof %d eoh %d  sd->sof %d\n", len, sof, eoh, sd->sof);
			gspca_dev->last_packet_type = DISCARD_PACKET;	
			sd->sof = 0;
			sd->framesize = 0;
			return;

		}
}

static int sd_querymenu(struct gspca_dev *gspca_dev,
                        struct v4l2_querymenu *menu)
{
        switch (menu->id) {

		case V4L2_CID_USER_CLASS:
                switch (menu->index) {
			/* Video Signal Source */
                	case 0:         
                        	strcpy((char *) menu->name, "CVBS");
                        	return 0;
                	case 1:        
                        	strcpy((char *) menu->name, "Y/C");
                        	return 0;
                }
        	break;
	}
        return -EINVAL;
}

static int sd_setvideosignalsource(struct gspca_dev *gspca_dev, __s32 val)
{
	struct sd *sd = (struct sd *) gspca_dev;

//	printk("sd_setvideosignalsource(%d)\n", val);

	i2c_register_write(gspca_dev, CAM_ADDR, V_SIGNAL, val);
 	sd->videosignalsource = val;
	
	i2c_register_write(gspca_dev, CAM_ADDR, 0x50, 0xff);
	return 0;
}

static int sd_getvideosignalsource(struct gspca_dev *gspca_dev, __s32 *val)
{
	struct sd *sd = (struct sd *) gspca_dev;
	
	*val = (u32)sd->videosignalsource;
//	printk("sd_getvideosignalsource() = %d\n", *val);

	return 0;
}

static int sd_setgain(struct gspca_dev *gspca_dev, __s32 val)
{
	struct sd *sd = (struct sd *) gspca_dev;
        
//	printk("sd_setgain(%d)\n", val);
	
	//i2c_register_write(gspca_dev, CAM_ADDR, ANALOG_GAIN, val);
	i2c_register_write(gspca_dev, CAM_ADDR, EEPROM_PAGE_1_ADDR + ANALOG_GAIN, val);
	sd->gain = val;
	mdelay(30);

	return 0;
}

static int sd_getgain(struct gspca_dev *gspca_dev, __s32 *val)
{
	struct sd *sd = (struct sd *) gspca_dev;

	*val = (u32)sd->gain;
//	printk("sd_getgain() = %d\n", *val);


	return 0;
}

static int sd_sethue(struct gspca_dev *gspca_dev, s32 val)
{
	struct sd *sd = (struct sd *) gspca_dev;

//	printk("sd_sethue(%d)\n", val);

	/* writing to eeprom addr seems to work better than the real reg addr */
	//i2c_register_write(gspca_dev, CAM_ADDR, HUE, val);
	i2c_register_write(gspca_dev, CAM_ADDR, EEPROM_PAGE_1_ADDR + HUE, val);
	sd->hue = val;
	mdelay(30);

	return 0;
}

static int sd_gethue(struct gspca_dev *gspca_dev, __s32 *val)
{
	struct sd *sd = (struct sd *) gspca_dev;
	
	*val = (u32)sd->hue;
//	printk("sd_gethue() = %d\n", *val);

	return 0;
}

static int sd_setchromagain(struct gspca_dev *gspca_dev, s32 val)
{
	struct sd *sd = (struct sd *) gspca_dev;

//	printk("sd_setchromagain(%d)\n", val);
	
	/* writing to eeprom addr seems to work better than the real reg addr */
	//i2c_register_write(gspca_dev, CAM_ADDR, CHROMA, val);
	i2c_register_write(gspca_dev, CAM_ADDR, EEPROM_PAGE_1_ADDR + CHROMA, val);
	sd->chromagain = val;
	mdelay(30);

	return 0;
}

static int sd_getchromagain(struct gspca_dev *gspca_dev, __s32 *val)
{
	struct sd *sd = (struct sd *) gspca_dev;

	*val = (u32)sd->chromagain;
//	printk("sd_getchromagain() = %d\n", *val);

	return 0;
}

static int sd_setcontrast(struct gspca_dev *gspca_dev, s32 val)
{
	struct sd *sd = (struct sd *) gspca_dev;

//	printk("sd_setcontrast(%d)\n", val);

	/* writing to eeprom addr seems to work better than the real reg addr */
	//i2c_register_write(gspca_dev, CAM_ADDR, CONTRAST, val);
	i2c_register_write(gspca_dev, CAM_ADDR, EEPROM_PAGE_1_ADDR + CONTRAST, val);
	sd->contrast = val;
	mdelay(30);

	return 0;
}

static int sd_getcontrast(struct gspca_dev *gspca_dev, __s32 *val)
{
	struct sd *sd = (struct sd *) gspca_dev;

	*val = (u32)sd->contrast;
//	printk("sd_getcontrast() = %d\n", *val);

	return 0;
}

static int sd_setbrightness(struct gspca_dev *gspca_dev, s32 val)
{
	struct sd *sd = (struct sd *) gspca_dev;

//	printk("sd_setbrightness(%d)\n", val);

	/* writing to eeprom addr seems to work better than the real reg addr */
	//i2c_register_write(gspca_dev, CAM_ADDR, BRIGHTNESS, val);
	i2c_register_write(gspca_dev, CAM_ADDR, EEPROM_PAGE_1_ADDR + BRIGHTNESS, val);
	sd->brightness = val;
	mdelay(30);

	return 0;
}

static int sd_getbrightness(struct gspca_dev *gspca_dev, __s32 *val)
{
	struct sd *sd = (struct sd *) gspca_dev;

	*val = (u32)sd->brightness;
//	printk("sd_getbrightness() = %d\n", *val);

	return 0;
}

static int sd_setsaturation(struct gspca_dev *gspca_dev, __s32 val)
{
	struct sd *sd = (struct sd *) gspca_dev;

//	printk("sd_setsaturation(%d)\n", val);

	/* writing to eeprom addr seems to work better than the real reg addr */
	//i2c_register_write(gspca_dev, CAM_ADDR, SATURATION, val);
	i2c_register_write(gspca_dev, CAM_ADDR, EEPROM_PAGE_1_ADDR + SATURATION, val);
	sd->saturation = val;
	mdelay(30);

        return 0;
}

static int sd_getsaturation(struct gspca_dev *gspca_dev, __s32 *val)
{
	struct sd *sd = (struct sd *) gspca_dev;
	
	*val = (u32)sd->saturation;
//	printk("sd_getsaturation() = %d\n", *val);

        return 0;
}

/* sub-driver description */
static const struct sd_desc sd_desc = {
	.name = MODULE_NAME,
	.ctrls = sd_ctrls,
	.nctrls = ARRAY_SIZE(sd_ctrls),
	.config = sd_config,
	.init = sd_init,
	.start = sd_start,
	.stop0 = sd_stop0,
	.stopN = sd_stopN,
	.dq_callback = dq_callback,
	.pkt_scan = sd_pkt_scan,
	.querymenu = sd_querymenu,
};

/* -- module initialisation -- */
static const __devinitdata struct usb_device_id device_table[] = {
	{USB_DEVICE(USB_VIDEOLOGY_VENDOR_ID, USB_VIDEOLOGY_PRODUCT_ID_FG),	/*.driver_info = SENSOR_NTSC*/}, /* framegrabber */
	{}
};
MODULE_DEVICE_TABLE(usb, device_table);

/* -- device connect -- */
/* need __devinit for kernels >= 2.6.33 */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,33)
static int __devinit sd_probe(struct usb_interface *intf,
			const struct usb_device_id *id)
#else
static int sd_probe(struct usb_interface *intf,
			const struct usb_device_id *id)
#endif
{
//	printk("sd_probe()\n");
	return gspca_dev_probe(intf, id, &sd_desc, sizeof(struct sd),
				THIS_MODULE);
}

static struct usb_driver sd_driver = {
	.name = MODULE_NAME,
	.id_table = device_table,
	.probe = sd_probe,
	.disconnect = gspca_disconnect,
#ifdef CONFIG_PM
	.suspend = gspca_suspend,
	.resume = gspca_resume,
#endif
};

/* -- module insert / remove -- */
static int __init sd_mod_init(void)
{
	int ret;

	//printk("sd_mod_init()\n");
	ret = usb_register(&sd_driver);
	if (ret < 0)
		return ret;
	PDEBUG(D_PROBE, "registered");
	return 0;
}
static void __exit sd_mod_exit(void)
{
	//printk("sd_mod_exit()\n");
	usb_deregister(&sd_driver);
	PDEBUG(D_PROBE, "deregistered");
}

module_init(sd_mod_init);
module_exit(sd_mod_exit);

/* Work around for selecting format size through viewer. */
/* Use this param if you want to force the viewer to a specific */
/* format size. This will override any height/width requested by
/* the viewer or specified on viewer command line. */ 
module_param(default_mode, int, 0644);
MODULE_PARM_DESC(default_mode, "default_mode: Default format size (0=viewer selects), (1=PAL 724x486), (2=NTSC 724x482), (3=NTSC 724x576)");


/* Offset (skip) the first N bytes of the frame. This is needed 
   to work around issues with some viewers.  If you have funny colors, 
   try a frame offset value between 1-3. 

   This module_param can also be used to skip over the header if
   needed (with some cameras), so as to align with the start of active video.   
*/
module_param(frame_offset, int, 0644);
MODULE_PARM_DESC(frame_offset, "frame_offset: Number of bytes to skip over to align the SOF with the first byte of the first pixel.");

